// Unit.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;

/**
 * The Unit class instances are micro unit in our virtual enterprise;
 * i.e. the units where the steps required to build a product (to fulfill
 * an order) are done
 *
 * @author Pietro Terna (with Marco Remondino, Dario Landini, Nicola 
 * Ormezzano and Paolo Pelligra contributions)
 */
public class Unit extends SwarmObjectImpl{

    /** the number identifying the unit*/
    public int unitNumber;

    /** the number of different production phases the unit is able to 
     * deal with (the 0 copy of this field is used for technical reasons) */
    public int nOfPhasesToDealWith=1, nOfPhasesToDealWith0=0, countPhases=0;
    /** the unit production phase */
    public int[] unitProductionPhases;




    /** the current state of production phase (what are we doing) 
     *  
     *  the current set is established by the order currently produced 
     *  by this unit in this tick
     * */
    public int currentState=-1; // -1 stays for undefined state

    /** choosing if we use or not the warehouses
	(to provide or not inventories of our products) */
    public boolean useWarehouses;
    /** choosing if we use or not newses about incoming productions*/
    public boolean useNewses;
    /** a flag about using pieces in warehouse */
    public boolean madeProduction;
    /** the list of all the units existing in our environment*/
    public ListImpl unitList;
    /** the list of the orders to be executed */
    public ListImpl  waitingList;
    /** the list containing the orders (if any) performed in a day, also
	via the inventories contained in our warehouse */
    public ListImpl dailyProductionList;
    /** the modelSwarm (to which we apply the 
     * setProductionTimeAndRecipeLength(...) method */
    VEFrameModelSwarm vEFrameModelSwarm;
    /** the temporary address of an existing unit */
    public Unit aUnit;
     /** the temporary address of a unit to wich send a news */
    public Unit receivingUnit;
    /** our warehouse, if any */
    public Warehouse myWarehouse;
    /** the first order to be executed */
    public Order firstOrder;
    /** an order inserted in a news */
    public Order incomingOrder;
    /** a news to be sent */
    public News aNews;
    /** a news to be deleted a the arrive of the related order */
    public News toBeDroppedNews;
    /** our rule master about inventories */
    public InventoryRuleMaster inventoryRuleMaster;
     /** our rule master about information system */
    public InformationRuleMaster informationRuleMaster;
    /** the list of the newses */
    public ListImpl newsList;
    /** how deep we look at production steps to propagate news (how many steps 
     * beyond the current production phase)*/
    public int informationDeepness;
    /** of order for which newses are propagated */
    public int nOfOrdersInNewses;
    /** the whole list of the orders */
    public ListImpl wholeOrderList;

    /** the number of different values of fixed and variable costs,
     *  related to the number of productin phases production phases */
    public int nOfCostTypes=1, nOfCostTypes0=0, countCostTypes=0;
    /** our fixed and variable costs */
    public float[] fixedCosts, variableCosts;

    /** total tick variable costs */
    public float totalTickVariableCosts;
    /** total daily fixed and variable costs */
    public float totalDailyCosts;
    /** the costs from the beginning of the simulation */
    public float totalCosts = 0;
    /** the daily revenues from the finished orders (sold) */
    public float dailySaleRevenues;
    /** the revenues from the beginning of the simulation */
    public float totalSaleRevenues = 0;
    /** the costs of the finished orders */
    public float totalFinishedOrderCosts = 0;
    /** the daily evaluation of the stored component */
    public float dailyStoredComponentValue = 0;
    /** the cumulative inventory financial costs generated from inventories */
    float totalInventoryFinancialCosts = 0;

    /**
     * the first constructor for Unit (using warehouses)
     */
    public Unit (Zone aZone, ListImpl ul, VEFrameModelSwarm mo,
		 InventoryRuleMaster rm, Warehouse wh, ListImpl ol)
    {
	// Call the constructor for the Unit parent class.
	super(aZone);

	useWarehouses = true;      // if we are using this constructor,
	                           // warehouses are activated
	useNewses = false;         // if we are using this constructor,
	                           // newses are not activated
        inventoryRuleMaster = rm;  // our InventoryRuleMaster
	myWarehouse = wh;	   // our warehouse address

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// the list of all the units
	unitList=ul;
	// the model
	vEFrameModelSwarm = mo;
	// the whole list of orders
        wholeOrderList = ol;
	// the daily list of finished orders
	dailyProductionList= new ListImpl(getZone());

	// announce the unit to the console
	if(StartVEFrame.verbose)
	    System.out.println("Unit number " + unitNumber + " has been created.");
    }

    /**
     * the second constructor for Unit (without warehouses)
     */
    public Unit (Zone aZone, ListImpl ul, VEFrameModelSwarm mo,
                  ListImpl ol) {

	// Call the constructor for the Unit parent class
	super(aZone);

	useWarehouses = false;      // if we are using this constructor,
	                            // warehouses are not activated
	useNewses = false;          // if we are using this constructor,
	                            // newses are not activated
        inventoryRuleMaster = null; // to avoid compiler warning/error
	myWarehouse = null;         // " " "

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// the list of all the units
	unitList=ul;
	// the model
	vEFrameModelSwarm = mo;
	// the daily list of finished orders
	dailyProductionList= new ListImpl(getZone());
	// the whole list of orders
        wholeOrderList = ol;

	// announce the unit to the console
	if(StartVEFrame.verbose)
	    System.out.println("Unit number " + unitNumber + " has been created.");
    }

    /**
     * the third constructor for Unit (with warehouses and newses)
     */
    public Unit (Zone aZone, ListImpl ul, VEFrameModelSwarm mo,
		 InventoryRuleMaster rm, Warehouse wh, 
		 InformationRuleMaster irm, int idp, int noin, ListImpl ol) 
    {
	// Call the constructor for the Unit parent class.
	super(aZone);

	useWarehouses = true;       // if we are using this constructor,
	                            // warehouses are activated
        useNewses = true;           // if we are using this constructor,
	                            // newses are activated
	inventoryRuleMaster = rm;   // our InventoryRuleMaster
	informationRuleMaster= irm; // our InformationRuleMaster
        myWarehouse = wh;	    // our warehouse address

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// the list of all the units
	unitList=ul;
	// the model
	vEFrameModelSwarm = mo;
	// the daily list if finished orders
	dailyProductionList= new ListImpl(getZone());
        // the news list
        newsList= new ListImpl(getZone());
        // how many steps we consider in recipes to propagate newses
        informationDeepness= idp;
	// # of orders for which newses are propagated
	nOfOrdersInNewses = noin;    
	// the whole list of orders
        wholeOrderList = ol;

	// announce the unit to the console
	if(StartVEFrame.verbose)
	System.out.println("Unit number " + unitNumber + " has been created.");
    }

    /** set unit number */
    public void setUnitNumber(int un)
    {
	// the unit id number.
	unitNumber=un;
    }

    /** setting fixed and variable costs */
    public void setFixedAndVariableCosts(float fc, float vc)
    {
	nOfCostTypes=nOfPhasesToDealWith;
	if (nOfCostTypes != nOfCostTypes0)
	    {
		fixedCosts    = new float[nOfCostTypes];
		variableCosts = new float[nOfCostTypes];

		nOfCostTypes0 = nOfCostTypes;
		countCostTypes=0;
	    }
	// our fixed and variable costs
        fixedCosts[countCostTypes] = fc;
        variableCosts[countCostTypes] = vc;
	countCostTypes++;
    }

    /** zeroing daily costs */
    public void unitStep0 ()
    {
	// at the beginnig of the day we have not yet generated any cost or
	// revenue 
        totalDailyCosts = 0;
	dailySaleRevenues = 0;

	// choosing an undefined state
	currentState=-1;
    }

    /**
     * production:<br><br>
     *
     * at the beginning of each time tick in the 'day', each unit looks for the 
     * order(s) (if any) to be executed; the production cycle is 'one unit of time'
     * in a step of the recipe of an order in a tick of a day;<br><br>
     *
     * we can also use inventories
     *
     */
    public void unitStep1 ()
    {
	int i;
        // at the beginnig of the tick we have not yet generated variable costs
        totalTickVariableCosts = 0;

	if(StartVEFrame.verbose)
	    System.out.println("Step 1 in unit " + unitNumber +
			       "; waitingList count " + waitingList.getCount());

	// after the specific step in recipe production (directly or from 
	// inventories) we place the order(s) in a daily production
	// list, which will be used at unitStep2 to diffuse production
	// results

	// the production
        madeProduction=false;
	if(waitingList.getCount() > 0){
	    firstOrder=(Order)waitingList.removeFirst();

	    currentState=-1;
	    for (i=0; i<getNOfPhasesToDealWith() && currentState==-1;i++)
		if(firstOrder.getNextStep()==getProductionPhase(i))
		    currentState=i;

	    if(currentState==-1){
		System.out.println("Enrecoverable error in unit " + unitNumber);
		System.exit(1);
	    }

	    dailyProductionList.addLast(firstOrder);
	    if (useWarehouses && ! myWarehouse.empty()){
		myWarehouse.decreaseInventoryCounterValue();
	    }
	    else {
		madeProduction=true;
		addTickVariableCosts();
	    }
	    firstOrder.setDoneStep();
	    //set the costs of the order in production

	    firstOrder.setFixedAndVariableOrderCosts(
						     fixedCosts[currentState], 
						     variableCosts[currentState]);

        }

	// having made no production, check to increase inventories
	if(! madeProduction && useWarehouses){
	    if(useNewses){
		if(inventoryRuleMaster.increaseInventoriesUsingNewses
		   (unitNumber, myWarehouse, newsList)){
		    myWarehouse.increaseInventoryCounterValue();
		    //recording variable costs
/*!!*/		    addTickVariableCosts();
                }
	    }
	    else {
                if(inventoryRuleMaster.increaseInventories
                                          (unitNumber, myWarehouse)){
		    myWarehouse.increaseInventoryCounterValue();
		    //recording variable costs
/*!!!*/		    addTickVariableCosts();
                }
	    }
        }

	// the use of inventories
	if (useWarehouses)
	    {
		while(waitingList.getCount()>0 && ! myWarehouse.empty())
		    {
		    myWarehouse.decreaseInventoryCounterValue();
		    firstOrder=(Order)waitingList.removeFirst();
		    firstOrder.setDoneStep();
		      // set the production costs into the order
		      firstOrder.setFixedAndVariableOrderCosts(
					   fixedCosts[currentState], 
					   variableCosts[currentState]);
		    // is the next recipe step to be done in our unit? If yes, 
	            // assign the order immediately to ourselves (if there is 
	            // room in the inventories we can make the production step 
		    // in this same cycle); we manage this step in a FIFO
		    // way to assure continuity in production of a multi-tick
		    // step
/*??*/		     if(getProductionPhase(0)==firstOrder.getNextStep())
			 waitingList.addFirst(firstOrder);
		    // if not, put the order in the daily production list
		     else dailyProductionList.addLast(firstOrder);
		    }
	    }
    }

    /**
     * diffusion and accounting:<br><br>
     *
     * sending the orders to the subsequent production units;<br><br>
     *
     * making accounting operations
     */
    public void unitStep2 () {

	int i, ii;
        int x;
	boolean operatingUnitNotFound;

	if(StartVEFrame.verbose)
	    System.out.println("Step 2 in unit " + unitNumber +
			       " dailyProductionList count " +
			       dailyProductionList.getCount());

	while(dailyProductionList.getCount()>0)
	    {
		firstOrder=(Order)dailyProductionList.removeFirst();

		// do we have concluded the production step?
		// if yes ... we drop the object
		if (firstOrder.getNextStep()==0)
		    {
			if(StartVEFrame.verbose)
			    System.out.println("drop order # " +
					       firstOrder.getOrderNumber() +
					       " startingDay " +
					       firstOrder.getStartingDay() +
					       " endDay " +
					       Globals.env.getCurrentTime() +
					       " stepNumber " +
					       firstOrder.getStepNumber() +
					       " ratio " +
					       ( (float)
						(Globals.env.getCurrentTime() -
						firstOrder.getStartingDay()))/
					        firstOrder.getStepNumber());
			// the production is concluded, so we inform
			// our environment (the model) of the time spent
			// and of the # of steps in recipe
			vEFrameModelSwarm.setProductionTimeAndRecipeLength(
			  Globals.env.getCurrentTime() -
			  firstOrder.getStartingDay(),
			  firstOrder.getStepNumber());
			if(StartVEFrame.verbose)
			    System.out.println("Order #" +
					       firstOrder.orderNumber +
					       " has been completed");
			// remove the dropped order from the wholeOrderList
			wholeOrderList.remove(firstOrder);
                        // recording the revenue from the finished order
                        dailySaleRevenues += firstOrder.getOrderEvaluation();
                        //recording the costs of the order
                        totalFinishedOrderCosts += 
			    firstOrder.getOrderTotalCosts();
                        // now we drop the object (also to save memory space)
                        firstOrder.drop();
		    }

		// if not ... we send it to the subsequent production unit
		else
		    {
			operatingUnitNotFound=true;
			for (i=0;i<unitList.getCount() &&
				 operatingUnitNotFound;i++)
			    {
				aUnit=(Unit)unitList.atOffset(i);

//##
//## multiple phases in units and parallel steps in recipes
//## in units modify currentState

//@@
// confronti con 0.5.1 senza news e warehouses ma con current state attivato
				for (ii=0;ii<aUnit.getNOfPhasesToDealWith() &&
					 operatingUnitNotFound;ii++)
				    {
					if (firstOrder.getNextStep() ==
					    aUnit.getProductionPhase(ii))
					    {
						operatingUnitNotFound=false;
						aUnit.setWaitingList(firstOrder);
					    }
				    }
			    }
		    }
	    }

	// set accounting data
	setTotalDailyCosts();
	if(useWarehouses){
	    setTotalInventoryFinancialCosts();
	    setStoredComponentDailyValue();
	}

    }

    /*
     * setting daily accounting data (step4 is executed after step3, but
     * it is reported here for practical writing reasons */
    public void unitStep4 ()
    {
	setTotalCosts();
	setTotalSaleRevenues();
    }

    /*
     * dealing with newses
     */
    public void unitStep3 () {

	int n;
	int i;
	int m;
        int stepsToBeDone;
        int receivingUnitNumber;
	boolean operatingUnitNotFound;

	for(n=0; n<nOfOrdersInNewses && waitingList.getCount()>n && useNewses; n++) {
		// we get the first order from the waitingList to find how many steps 
		// are necessary to finish its recipe
		incomingOrder=(Order)waitingList.atOffset(n);
		stepsToBeDone=incomingOrder.getHowManyStepsToBeDone();

		// we ask the order about the subsequent production units, according 
		// information deepness (we start from next step, in position 2)
		for (i=2;i<=stepsToBeDone && i<=informationDeepness+1;i++)
		    {
			receivingUnitNumber = 
			    incomingOrder.getProductionToBeDoneAtStepAheadNumber(i);
			      if(StartVEFrame.verbose)
			          System.out.println(
                                        " Receiving Unit # " +
				        incomingOrder.
				        getProductionToBeDoneAtStepAheadNumber(i));
			if(informationRuleMaster.
			   checkToSendNews(unitNumber, receivingUnitNumber)
			   && unitNumber!= receivingUnitNumber)
			    {
				operatingUnitNotFound=true;
				// searching the units able to make the next production
				// steps and sending the newses
				for (m=0;
				     m<unitList.getCount() && operatingUnitNotFound;
				     m++)
				    {
					receivingUnit=(Unit)unitList.atOffset(m);
/*??*/	/*send to all units?*/		  if (receivingUnitNumber==receivingUnit.
					    getProductionPhase(0))
					    {
						aNews=new News (getZone(), unitNumber, 
								receivingUnit.
/*??*/ /* what here? */	 			       		getProductionPhase(0), 
								incomingOrder);
						receivingUnit.setNewsList(aNews);
						operatingUnitNotFound=false;
					    }
				    }
			    }
		    }
	    }
    }

    /** set unit production phases */
    public void setProductionPhase(int ph)
    {
	// the nOfPhasesToDealWith is 1 as default, but it can be newly 
	// defined if the unit is a special one
	if (nOfPhasesToDealWith != nOfPhasesToDealWith0)
	    {
		unitProductionPhases = new int[nOfPhasesToDealWith]; 

		nOfPhasesToDealWith0 = nOfPhasesToDealWith;
		countPhases=0;
	    }
	unitProductionPhases[countPhases] = ph;
	countPhases++;
    }

    /**
     * return the production phases of the unit
     */
    public int getProductionPhase(int state)
    {
	return unitProductionPhases[state]; 
    }

    /**
     * Putting an order in the production waitingList of the unit
     * and deleting the related news
     */
    public void setWaitingList(Order anOrder)
    {
       boolean newsToBeDroppedNotFound;
       int i; 

        waitingList.addLast(anOrder);
	if(StartVEFrame.verbose)
	    System.out.println("The unit " + unitNumber +
			       " has received the order # "
			       + anOrder.getOrderNumber());

	if (useNewses && newsList.getCount()>0)
	    //looking for a news containing information about
	    // the order just arrived to the unit
	    for (i=1;i<=newsList.getCount();i++)
		{
		    toBeDroppedNews=(News)newsList.removeFirst();
		    if (toBeDroppedNews.incomingOrderInNews.getOrderNumber()
			==anOrder.getOrderNumber())
			{
			    toBeDroppedNews.drop();
			    if(StartVEFrame.verbose)
				System.out.println(
					"Dropping the news announcinig the order n "
					+ toBeDroppedNews.incomingOrderInNews.
					                     getOrderNumber());
			}
		    else
			newsList.addLast(toBeDroppedNews);
		}
    }

    /**
     * return the waiting list length
     */
    public int getWaitingListLength()
    {
	return waitingList.getCount();
    }

    /**
     * putting a news in the newsList of the unit after
     * checking list to avoid duplicated news
     */
    public void setNewsList(News aNews)
    {
	int i;
	boolean  duplicatedNews;
	
	// a single news of the newsList
	News controlNews;
	// the news just arrived
	News toBeCheckedNews;

	toBeCheckedNews=aNews;
	duplicatedNews=false;

	if (newsList.getCount()>0)
	    {
		for (i=0;i<newsList.getCount() && ! duplicatedNews;i++)
		    {
			controlNews=(News) newsList.removeFirst();
			newsList.addLast(controlNews);
			if (toBeCheckedNews.incomingOrderInNews.getOrderNumber()==
			    controlNews.incomingOrderInNews.getOrderNumber()
			    && toBeCheckedNews.getNumberOfTheStepForWhichANewsIsSent()==
			    controlNews.getNumberOfTheStepForWhichANewsIsSent())

			    duplicatedNews=true;
		    }

		if (! duplicatedNews)
		    {
                  newsList.addLast(toBeCheckedNews);
                  if(StartVEFrame.verbose)
		      System.out.println("Unit " + unitNumber +
					 " accept a news containing order # "
					 + toBeCheckedNews.incomingOrderInNews.
					 getOrderNumber());
		    }
		else
		    if(StartVEFrame.verbose)
			System.out.println("Unit " + unitNumber +
					   " doesn't accept a news containing order # "
					   + toBeCheckedNews.incomingOrderInNews.
					   getOrderNumber()+
					   " which just exists for the same step");
            }
	else
	    {
		newsList.addLast(toBeCheckedNews);
		if(StartVEFrame.verbose)
		    System.out.println("Unit " + unitNumber +
				       " accepts a news containing order N "
				       + toBeCheckedNews.incomingOrderInNews.
				       getOrderNumber()+
				       " because newsList is empty");
	    }
    }

    /** set the nOfPhasesToDealWith, i.e. the n. of different steps
     *  the unit is able to deal with */
    public void setNOfPhasesToDealWith(int ns)
    {
	nOfPhasesToDealWith=ns;
    }

    /** get the nOfPhasesToDealWith, i.e. the n. of different steps
     *  the unit is able to deal with */
    public int getNOfPhasesToDealWith()
    {
	return nOfPhasesToDealWith;
    }

    /** set the value of the total daily variable costs */
    public void addTickVariableCosts(){
        totalTickVariableCosts += variableCosts[currentState];
    }

    /** set the global sum of fixed (per day) and variable costs (per tick) 
     *  it can be calculated several times in a day, as a sequence of 
     *  partial values */
    public void setTotalDailyCosts(){
	float mean; int i;
	//if currentState=-1 (no production, so no state chosen in the unit)
	if(currentState == -1)
	    {
		mean=0;
		for (i=0; i<getNOfPhasesToDealWith();i++) mean+=fixedCosts[i];
		mean /= getNOfPhasesToDealWith();
		totalDailyCosts += mean + totalTickVariableCosts;
	    }
	else
	    totalDailyCosts += 
		fixedCosts[currentState] + totalTickVariableCosts;
    }

    /** cumulate the total daily costs 
     *  it must be executed only once a day*/
    public void setTotalCosts(){
      totalCosts += totalDailyCosts;
    }

    /** return total daily costs */
    public float getTotalDailyCosts(){
      return totalDailyCosts;
    }

    /** return total costs */
    public float getTotalCosts(){
      return totalCosts;
    }

    /** record the revenue from a dropped order */
    public float getDailySaleRevenues(){
      return dailySaleRevenues;
    }

    /** cumulatere the revenues from finished orders */
    public void setTotalSaleRevenues(){
      totalSaleRevenues += dailySaleRevenues;
    }

    /** return the total revenue */
    public float getTotalSaleRevenues(){
      return totalSaleRevenues;
    }

    /** return total costs from finished orders */
    public float getTotalFinishedOrderCosts(){
	return totalFinishedOrderCosts;
    }

    /** record the cumulated financial costs of inventories */
    public void setTotalInventoryFinancialCosts(){
        totalInventoryFinancialCosts += (myWarehouse.getInventoryQuantity()*
             (vEFrameModelSwarm.getInventoryFinancialRate()/200));
	        // getInventoryFinancialRate returns an yearly rate, which is here
	        // divided by 200 day of labour per year in an approximate way
    }

    /** return the financial costs of inventories */
    public float getTotalInventoryFinancialCosts(){
        return totalInventoryFinancialCosts;
    }

    /** record the daily value of inventories */
    public void setStoredComponentDailyValue(){
        if(vEFrameModelSwarm.getInventoryEvaluationCriterion()==3)
	    dailyStoredComponentValue = myWarehouse.getInventoryQuantity()*
		vEFrameModelSwarm.getRevenuePerEachRecipeStep();

        else if(vEFrameModelSwarm.getInventoryEvaluationCriterion()==2)
	    dailyStoredComponentValue = myWarehouse.getInventoryQuantity()*
		(fixedCosts[currentState] + variableCosts[currentState]);

        else if(vEFrameModelSwarm.getInventoryEvaluationCriterion()==1)
	    dailyStoredComponentValue = myWarehouse.getInventoryQuantity()*
		variableCosts[currentState];

        else if(vEFrameModelSwarm.getInventoryEvaluationCriterion()!=1 &&
		vEFrameModelSwarm.getInventoryEvaluationCriterion()!=2 &&
		vEFrameModelSwarm.getInventoryEvaluationCriterion()!=3)
	    {
		System.out.println("your criterion doesn't exist: the variable"+
				   " inventoryEvaluationCriterion must 1, 2 or 3");
		System.exit(1);
	    }
    }

    public float getDailyStoredComponentValue(){
        return dailyStoredComponentValue;
    }

}



